/*
 * kgraft_patch_bsc983144
 *
 * Fix for CVE-2016-1583, bsc#983144
 *
 *  SLE12 commit:
 *  754f7d2e4ad53233beac042837b7d5ac39ca67d0
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/fs.h>
#include <linux/jump_label.h>

#include <linux/ratelimit.h>

#define KGR_PATCHED_MODULE      "ecryptfs"

static struct vm_operations_struct *kgr_generic_file_vm_ops;
static struct file_operations *kgr_ecryptfs_main_fops;

static struct kgr_kallsyms {
	char *name;
	void **addr;
} kgr_funcs_vmlinux[] = {
	{ "generic_file_vm_ops", (void *)&kgr_generic_file_vm_ops },
}, kgr_funcs_ecryptfs[] = {
	{ "ecryptfs_main_fops", (void *)&kgr_ecryptfs_main_fops },
};

struct static_key kgr_ecryptfs_key = STATIC_KEY_INIT_FALSE;

struct ecryptfs_dentry_info {
	struct path lower_path;
	struct ecryptfs_crypt_stat *crypt_stat;
};

static inline struct dentry *kgr_ecryptfs_dentry_to_lower(struct dentry *dentry)
{
	return ((struct ecryptfs_dentry_info *)dentry->d_fsdata)->lower_path.dentry;
}

/* patched function */
int kgr_generic_file_mmap(struct file * file, struct vm_area_struct * vma)
{
	struct address_space *mapping = file->f_mapping;
	/*
	 * Fix CVE-2016-1583
	 *  +next line
	 */
	struct dentry *dentry;

	/*
	 * Fix CVE-2016-1583
	 *  +next static_key part
	 */
	if (static_key_false(&kgr_ecryptfs_key)) {
		if (file->f_op == kgr_ecryptfs_main_fops) {
			dentry = kgr_ecryptfs_dentry_to_lower(file->f_dentry);

			if (!dentry->d_inode->i_fop->mmap)
				return -ENODEV;
		}
	}

	if (!mapping->a_ops->readpage)
		return -ENOEXEC;
	file_accessed(file);
	vma->vm_ops = kgr_generic_file_vm_ops;
	return 0;
}

static int kgr_patch_bsc983144_kallsyms(struct kgr_kallsyms *kgr_funcs, int length)
{
	unsigned long addr;
	int i;

	for (i = 0; i < length; i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc983144_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if ((action != MODULE_STATE_COMING && action != MODULE_STATE_GOING) ||
	     strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	if (action == MODULE_STATE_GOING) {
		static_key_slow_dec(&kgr_ecryptfs_key);
		return 0;
	}

	static_key_slow_inc(&kgr_ecryptfs_key);

	ret = kgr_patch_bsc983144_kallsyms(kgr_funcs_ecryptfs,
		ARRAY_SIZE(kgr_funcs_ecryptfs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc983144_module_nb = {
	.notifier_call = kgr_patch_bsc983144_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc983144_init(void)
{
	int ret;

	ret = kgr_patch_bsc983144_kallsyms(kgr_funcs_vmlinux,
		ARRAY_SIZE(kgr_funcs_vmlinux));
	if (ret)
		return ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc983144_kallsyms(kgr_funcs_ecryptfs,
			ARRAY_SIZE(kgr_funcs_ecryptfs));
		if (ret)
			goto out;

		static_key_slow_inc(&kgr_ecryptfs_key);
	}

	ret = register_module_notifier(&kgr_patch_bsc983144_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc983144_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc983144_module_nb);
}
